#pragma once

#include <libc/component.h>
#include <arpa/inet.h>
#include <errno.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <unistd.h>
#include <netdb.h>
#include <util/list.h>
#include <base/heap.h>
#include <picosha2/picosha2.h>
#include <terminal_session/connection.h>
#include "command.h"
#include "error.h"
#include "exception.h"
#include "payload.h"
#include "message.h"
#include "socket_wrapper.h"

#include <util/hexdump.h>

namespace Hsm {
  class Connection;

	using Hostname = Genode::String<256>;
	using Ip_address = Genode::String<64>;
	using Username = Genode::String<256>;
	using Password = Genode::String<256>;
	using Key_name = Genode::String<256>;
	using Crypto_method_label = Genode::String<256>;

	Crypto_method method_from_label(const Crypto_method_label label)
	{
		if (label == Crypto_method_label("EcDsaNone")) {
			return Crypto_method::EcDsaNone;
		} else {
			Genode::error("Unknown crypto method ", label);
			throw Invalid_argument_exception(); 
		}
	}
}

class Hsm::Connection
{
	private:
		enum {
			INVALID_SOCKET = -1,
			SALT_SIZE = 32,
			HASH_SIZE = 32,
			MAGIC_SIZE = 32,
		};

		Genode::Env&           _env;
		Genode::Heap& _heap;

		Terminal::Connection& _random_terminal;
		const Username         _username;
		const Password         _password;
		mutable bool           _login_success    { false };
		Socket_wrapper         _socket;

		Ip_address resolve_hostname(const Hostname& hostname) const
		{
			struct hostent *host_entry;
			struct in_addr **addr_list;
			int i;

			if ((host_entry = gethostbyname(hostname.string())) == NULL) {
				Genode::error("gethostbyname for ", hostname, " failed");
				throw Genode::Exception();
			}

			addr_list = (struct in_addr **)host_entry->h_addr_list;

			for(i = 0; addr_list[i] != NULL; i++) {
				return Ip_address(static_cast<const char*>(inet_ntoa(*addr_list[i])));
			}

			Genode::error("no ip address for ", hostname, " found");
			throw Genode::Exception();
		}

		void _login_request(const Username& username,
		                    uint8_t* client_login_salt,
		                    uint8_t* server_login_salt) const
		{
			Message request(_heap, Command::CommandLoginRequest);
			request.add(Data_type::typeIdLabel, username.string());
			_random_terminal.read(client_login_salt, SALT_SIZE);
			request.add(Data_type::typeRandomData, client_login_salt, SALT_SIZE);
			request.add(Data_type::typeUserId, username.string());

			Message& response = execute(request);

			switch (response.status())
			{
				case CKR_OK:
					break;
				default:
					Genode::error("HSM replied with error ", Genode::Hex(response.status()));
					throw Hsm::Error_response_exception();
			}

			auto server_salt_payload = response.payload_by_type(Data_type::typeLoginData);
			if (server_salt_payload == nullptr) {
				Genode::error("Server salt missing");
				throw Missing_payload_exception();
			}
			Genode::memcpy(server_login_salt, server_salt_payload->payload_buffer(), SALT_SIZE);
			destroy(_heap, &response);
		}

		void _compute_challenge(const uint8_t* magic,
		                        const uint8_t* salt1,
		                        const uint8_t* salt2,
		                        const Password& password,
		                        uint8_t* challenge) const
		{
			const char* const pw = password.string();
			std::vector<uint8_t> pw_hash(HASH_SIZE);
			picosha2::hash256(pw, pw + strlen(pw), pw_hash.begin(), pw_hash.end());

			picosha2::hash256_one_by_one hasher;
			hasher.process(magic, magic + MAGIC_SIZE);
			hasher.process(salt1, salt1 + SALT_SIZE);
			hasher.process(salt2, salt2 + SALT_SIZE);
			hasher.process(pw_hash.begin(), pw_hash.end());
			hasher.finish();

			hasher.get_hash_bytes(challenge, challenge + HASH_SIZE);
		}

		bool _login_response(const Password& password,
		                     const uint8_t* client_login_salt,
		                     const uint8_t* server_login_salt) const
		{
			uint8_t request_hash[HASH_SIZE];
			_compute_challenge(magic_login_request, server_login_salt,
			                   client_login_salt, password, request_hash);

			Message request(_heap, Command::CommandLoginResponse);
			request.add(Data_type::typeLoginData, request_hash, sizeof(request_hash));
			uint8_t client_auth_salt[SALT_SIZE];
			_random_terminal.read(client_auth_salt, sizeof(client_auth_salt));
			request.add(Data_type::typeRandomData, client_auth_salt, sizeof(client_auth_salt));

			Message& response = execute(request);

			switch (response.status())
			{
				case CKR_OK:
					break;
				default:
					Genode::error("HSM replied with error ", Genode::Hex(response.status()));
					throw Hsm::Error_response_exception();
			}

			auto server_salt_payload = response.payload_by_type(Data_type::typeRandomData);
			if (server_salt_payload == nullptr) {
				Genode::error("Server salt missing");
				throw Missing_payload_exception();
			}
			auto client_challenge = response.payload_by_type(Data_type::typeLoginData);
			if (client_challenge == nullptr) {
				Genode::error("Client_challenge missing");
				throw Missing_payload_exception();
			}
			uint8_t auth_hash[HASH_SIZE];
			_compute_challenge(magic_login_auth, server_salt_payload->payload_buffer(),
			                   client_auth_salt, password, auth_hash);
			bool login_successful = (Genode::memcmp(auth_hash, client_challenge->payload_buffer(), SALT_SIZE) == 0);

			destroy(_heap, &response);
			return login_successful;
		}

	public:

		Connection(Genode::Env& env,
		           Genode::Heap& heap,
		           Terminal::Connection& random_terminal,
		           const Ip_address& ip_address,
		           const uint16_t port,
		           const Username& username,
		           const Password& password) :
			_env             { env },
			_heap            { heap },
			_random_terminal { random_terminal },
			_username        { username },
			_password        { password },
			_socket          { env, ip_address, port,
			                   [&] () { _connected(); },
			                   [&] () { _disconnected(); } }
		{ }

		Message& execute(const Message& request) const
		{
			using Genode::Hex;
			request.write_to_socket(_socket);
			auto response = new (_heap) Message(_heap, _socket);

			if (response->status() >= SECUROSYS_ERROR_BASE) {
				Genode::error("HSM replied with error ", Genode::Hex(response->status()));
				throw Hsm::Error_response_exception();
			}

			return *response;
		}

		void hello() const
		{
			Message request(_heap, Command::CommandHello);
			request.add(Data_type::typeMagic, LOGIN_MAGIC);

			Message& response = execute(request);

			switch (response.status())
			{
				case CKR_OK:
					break;
				default:
					Genode::error("HSM replied with error ", Genode::Hex(response.status()));
					throw Hsm::Error_response_exception();
			}

			destroy(_heap, &response);
		}

		bool login(const Username& username, const Password& password) const
		{
			uint8_t client_salt[32];
			uint8_t server_salt[32];
			_login_request(username, client_salt, server_salt);
			return _login_response(password, client_salt, server_salt);
		}

		void _disconnected()
		{
			Genode::log("HSM disconnected");
			_login_success = false;
		}

		void _connected()
		{
			Genode::log("HSM connected");
			hello();
			_try_login();
		}

		bool _try_login() const
		{
			if (_login_success) {
				return _login_success;
			}

			if (login(_username, _password)) {
				_login_success = true;
				log("HSM login successful.");
			} else {
				error("HSM login failed.");
			}

			return _login_success;
		}

		uint32_t sign(const Crypto_method method,
		              const Key_name& key_name,
		              const uint8_t* message_data,
		              const uint32_t message_size,
		              uint8_t* signature_data,
		              const uint32_t signature_size) const
		{
			Message request(_heap, Command::CommandAsymmetricSingleSign);
			request.add(Data_type::typeCryptoMethod, static_cast<uint32_t>(method));
			request.add(Data_type::typeIdLabel, key_name.string());
			request.add(Data_type::typeSignData, message_data, message_size);

			Message& response = execute(request);

			switch (response.status())
			{
				case CKR_OK:
					break;
				default:
					Genode::error("HSM replied with error ", Genode::Hex(response.status()));
					throw Hsm::Error_response_exception();
			}

			auto signature = response.payload_by_type(Data_type::typeSignatureData);
			if (signature == nullptr) {
				Genode::error("Signature data missing");
				throw Missing_payload_exception();
			}
			if (signature_size < signature->payload_length()) {
				Genode::error("Signature buffer to small");
				throw Buffer_too_short_exception();
			}
			const uint32_t output_size = signature->payload_length();
			Genode::memcpy(signature_data, signature->payload_buffer(), output_size);

			destroy(_heap, &response);

			return output_size;
		}

		bool verify(const Crypto_method method,
		            const Key_name& key_name,
		            const uint8_t* message_data,
		            const uint32_t message_size,
		            const uint8_t* signature_data,
		            const uint32_t signature_size) const
		{
			Message request(_heap, Command::CommandAsymmetricSingleVerify);
			request.add(Data_type::typeCryptoMethod, static_cast<uint32_t>(method));
			request.add(Data_type::typeIdLabel, key_name.string());
			request.add(Data_type::typeVerifyData, message_data, message_size);
			request.add(Data_type::typeSignatureData, signature_data, signature_size);

			Message& response = execute(request);

			switch (response.status())
			{
				case CKR_OK:
					break;
				default:
					Genode::error("HSM replied with error ", Genode::Hex(response.status()));
					throw Hsm::Error_response_exception();
			}

			bool signature_good = false;

			switch (response.status())
			{
				case CKR_OK:
					signature_good = true;
					break;
				case CKR_SIGNATURE_INVALID:
					signature_good = false;
					break;
				default:
					Genode::error("HSM replied with error ", Genode::Hex(response.status()));
					throw Hsm::Error_response_exception();
			}

			destroy(_heap, &response);

			return signature_good;
		}
};
